@@ -41,8 +41,9 @@ enum VectorType nanoarrow_infer_vector_type_array(SEXP array_xptr);
 // dispatch to find a convert_array() method (or error if there
 // isn't one)
 static SEXP call_convert_array(SEXP array_xptr, SEXP ptype_sexp) {
-  SEXP fun = PROTECT(Rf_install("convert_array_from_c"));
-  SEXP call = PROTECT(Rf_lang3(fun, array_xptr, ptype_sexp));
+  SEXP fun = PROTECT(Rf_install("convert_fallback_other"));
+  // offset/length don't need to be modified in this case
+  SEXP call = PROTECT(Rf_lang5(fun, array_xptr, R_NilValue, R_NilValue, ptype_sexp));
   SEXP result = PROTECT(Rf_eval(call, nanoarrow_ns_pkg));
   UNPROTECT(3);
   return result;
@@ -99,31 +100,64 @@ static SEXP convert_array_default(SEXP array_xptr, enum VectorType vector_type,
   return result;
 }
 
-static SEXP convert_array_chr(SEXP array_xptr) {
-  SEXP result = PROTECT(nanoarrow_c_make_altrep_chr(array_xptr));
-  if (result == R_NilValue) {
-    call_stop_cant_convert_array(array_xptr, VECTOR_TYPE_CHR, R_NilValue);
-  }
-  UNPROTECT(1);
-  return result;
-}
-
-SEXP nanoarrow_c_convert_array(SEXP array_xptr, SEXP ptype_sexp);
-
-static SEXP convert_array_data_frame(SEXP array_xptr, SEXP ptype_sexp) {
-  // If array_xptr is a union, use default convert behaviour
+static SEXP convert_array_chr(SEXP array_xptr, SEXP ptype_sexp) {
   struct ArrowSchema* schema = schema_from_array_xptr(array_xptr);
   struct ArrowSchemaView schema_view;
   if (ArrowSchemaViewInit(&schema_view, schema, NULL) != NANOARROW_OK) {
     Rf_error("Invalid schema");
   }
 
-  if (schema_view.storage_type != NANOARROW_TYPE_STRUCT) {
-    ptype_sexp = PROTECT(nanoarrow_c_infer_ptype(array_xptr_get_schema(array_xptr)));
-    SEXP default_result =
-        convert_array_default(array_xptr, VECTOR_TYPE_DATA_FRAME, ptype_sexp);
+  // If array_xptr is an extension, use default conversion
+  if (schema_view.extension_name.size_bytes > 0) {
+    // Default conversion requires a ptype: resolve it if not already specified
+    if (ptype_sexp == R_NilValue) {
+      ptype_sexp = PROTECT(nanoarrow_c_infer_ptype(array_xptr_get_schema(array_xptr)));
+      SEXP default_result =
+          PROTECT(convert_array_default(array_xptr, VECTOR_TYPE_OTHER, ptype_sexp));
+      UNPROTECT(2);
+      return default_result;
+    } else {
+      return convert_array_default(array_xptr, VECTOR_TYPE_OTHER, ptype_sexp);
+    }
+  }
+
+  struct ArrowArray* array = (struct ArrowArray*)R_ExternalPtrAddr(array_xptr);
+  if (array->dictionary == NULL) {
+    SEXP result = PROTECT(nanoarrow_c_make_altrep_chr(array_xptr));
+    if (result == R_NilValue) {
+      call_stop_cant_convert_array(array_xptr, VECTOR_TYPE_CHR, R_NilValue);
+    }
     UNPROTECT(1);
-    return default_result;
+    return result;
+  } else {
+    return convert_array_default(array_xptr, VECTOR_TYPE_CHR, R_NilValue);
+  }
+}
+
+SEXP nanoarrow_c_convert_array(SEXP array_xptr, SEXP ptype_sexp);
+
+static SEXP convert_array_data_frame(SEXP array_xptr, SEXP ptype_sexp) {
+  struct ArrowSchema* schema = schema_from_array_xptr(array_xptr);
+  struct ArrowSchemaView schema_view;
+  if (ArrowSchemaViewInit(&schema_view, schema, NULL) != NANOARROW_OK) {
+    Rf_error("Invalid schema");
+  }
+
+  // If array_xptr is an extension, union, or the ptype isn't a data.frame
+  // use convert/materialize convert behaviour.
+  // Default conversion requires a ptype: resolve it if not already specified
+  if (schema_view.storage_type != NANOARROW_TYPE_STRUCT ||
+      schema_view.extension_name.size_bytes > 0 ||
+      (ptype_sexp != R_NilValue && !Rf_inherits(ptype_sexp, "data.frame"))) {
+    if (ptype_sexp == R_NilValue) {
+      ptype_sexp = PROTECT(nanoarrow_c_infer_ptype(array_xptr_get_schema(array_xptr)));
+      SEXP default_result =
+          PROTECT(convert_array_default(array_xptr, VECTOR_TYPE_OTHER, ptype_sexp));
+      UNPROTECT(2);
+      return default_result;
+    } else {
+      return convert_array_default(array_xptr, VECTOR_TYPE_DATA_FRAME, ptype_sexp);
+    }
   }
 
   struct ArrowArray* array = array_from_xptr(array_xptr);
@@ -184,7 +218,7 @@ SEXP nanoarrow_c_convert_array(SEXP array_xptr, SEXP ptype_sexp) {
       case VECTOR_TYPE_DBL:
         return convert_array_default(array_xptr, vector_type, R_NilValue);
       case VECTOR_TYPE_CHR:
-        return convert_array_chr(array_xptr);
+        return convert_array_chr(array_xptr, ptype_sexp);
       case VECTOR_TYPE_DATA_FRAME:
         return convert_array_data_frame(array_xptr, R_NilValue);
       default:
@@ -209,8 +243,9 @@ SEXP nanoarrow_c_convert_array(SEXP array_xptr, SEXP ptype_sexp) {
                Rf_inherits(ptype_sexp, "vctrs_list_of") ||
                Rf_inherits(ptype_sexp, "Date") || Rf_inherits(ptype_sexp, "hms") ||
                Rf_inherits(ptype_sexp, "POSIXct") ||
-               Rf_inherits(ptype_sexp, "difftime")) {
-      return convert_array_default(array_xptr, VECTOR_TYPE_OTHER, ptype_sexp);
+               Rf_inherits(ptype_sexp, "difftime") ||
+               Rf_inherits(ptype_sexp, "integer64")) {
+      return convert_array_default(array_xptr, VECTOR_TYPE_UNINITIALIZED, ptype_sexp);
     } else {
       return call_convert_array(array_xptr, ptype_sexp);
     }
@@ -225,7 +260,7 @@ SEXP nanoarrow_c_convert_array(SEXP array_xptr, SEXP ptype_sexp) {
     case REALSXP:
       return convert_array_default(array_xptr, VECTOR_TYPE_DBL, ptype_sexp);
     case STRSXP:
-      return convert_array_chr(array_xptr);
+      return convert_array_chr(array_xptr, ptype_sexp);
     default:
       return call_convert_array(array_xptr, ptype_sexp);
   }
