@@ -78,7 +78,11 @@ enum VectorType nanoarrow_infer_vector_type_schema(SEXP schema_xptr) {
     Rf_error("nanoarrow_infer_vector_type_schema(): %s", ArrowErrorMessage(&error));
   }
 
-  return nanoarrow_infer_vector_type(schema_view.type);
+  if (schema_view.extension_name.size_bytes > 0) {
+    return VECTOR_TYPE_OTHER;
+  } else {
+    return nanoarrow_infer_vector_type(schema_view.type);
+  }
 }
 
 // The same as the above, but from a nanoarrow_array()
